/* R. Harmsen, December 2023.
   First completely working version 15 December 2023. */

#include <stdlib.h>
#include "rhmidibendbuf.h"
#include "rhmidifuncs.h"
#include "rhmusicadhoc.h"

#define BENDBUFSIZE 40

enum BendBufReadStatus
{
   BENDBUF_2HDL = -3,
   BENDBUF_FILD = -2,
   BENDBUF_EOF  = -1,
   BENDBUF_FREE =  0,
};
struct MidiBendBufData
{
   enum   BendBufReadStatus Status;
   short  CorrespChordIndex;
   short  FrstCode;     /* 0x90 or 0xE0 */
   short  MidiChannel;  /* 0 ... 15 */
   short  MidiNote;     /* 0 ... 127 */
   short  Grado12;      /* The MidiNote relative to the tonica, ignoring the octave. */
   float  BendCents;
   long   EventFilepos; /* On the 1st byte of the 2-byte bend value */
   LLI    SigmaT;       /* In ticks */
   double SigmaQ;       /* In quarternotes */
};

static struct MidiBendBufData  BendEntry[BENDBUFSIZE];
static struct MidiBendBufData *BendPoitr[BENDBUFSIZE];

static int Chord[13];
#define NotesInChord ((int)((sizeof Chord / sizeof Chord[0])))

static struct DegreeData wrkC[NotesInChord];
static struct DegreeData resC[NotesInChord];

/* Encoded as a Midi note, derived from event FF-59. Default is central C, 60 */
static int Tonica;

/***********************************************************************
   Function prototypes.
 **********************************************************************/
static int  TrackCompare (const void *va, const void *vb);
static int  BendBufCompare (const void *va, const void *vb);
static int  AdhocWhichSlotToUse (void);
static void AdhocFillBenddata (int BufIndex, struct rhmidi_eventdata *EvtdatTrack);
static int  PickupTonality (struct rhmidi_eventdata *EvtdatTrack);
static void ShowBuf (int progline);

/***********************************************************************/
void AdhocRefillBendBuf (FILE *fpi, char *ProgPath, char *MidiFilename,
   struct rhmidi_eventdata **EvtdatTracks, int NumTracks,
   int TicksPerQuarter, int verbose)
{
   int BufIndex = 0, bi;
   int removedEn;

   do {
      while ((BufIndex = AdhocWhichSlotToUse()) >= 0)
      {
         if (EvtdatTracks[0]->EndOfTrack)
         {
            /* Let AdhocFillBenddata mark it in the buffer. */
            AdhocFillBenddata(BufIndex, EvtdatTracks[0]);
            break;
         }
         ReadMidiEventErrHandling(
            fpi, ProgPath, MidiFilename, EvtdatTracks[0], TicksPerQuarter, verbose);
         if (EvtdatTracks[0]->FrstCode == 0xFF &&
             EvtdatTracks[0]->ScndCode == 0x59)
         {
            Tonica = PickupTonality(EvtdatTracks[0]);
            if (verbose > 2)
               fprintf(stdout, "Line %4d, Tonica = %2d\n", __LINE__, Tonica);
         }
         else
         if ((EvtdatTracks[0]->FrstCode == 0xE0 &&
              CnvFmMidiBend(EvtdatTracks[0]->EventData) != 0.0)
              ||
             (EvtdatTracks[0]->FrstCode == 0x90 &&
              EvtdatTracks[0]->EventData[1] != 0))
         {
            AdhocFillBenddata(BufIndex, EvtdatTracks[0]);
         }
         /* Sort, so the pointer of the track, that the next event should be
            read from, is at the top of the array. */
         qsort(EvtdatTracks, NumTracks, sizeof *EvtdatTracks, TrackCompare);
      }

      /* The algorithm (in adhocmidi.c) for reading from the various
         tracks, does not always put the Midi events in the proper
         order. Therefore sort them here. */
      qsort(BendPoitr, BENDBUFSIZE, sizeof *BendPoitr, BendBufCompare);
      if (verbose > 2) ShowBuf(__LINE__);

      removedEn = 0;
      for (bi = 0; bi < BENDBUFSIZE - 1; bi++)
      {
         /* A non-resetting bend not followed immediately by a 'note on'?
            Perhaps a glissando. Ignore */
         if (BendPoitr[bi  ]->Status == BENDBUF_FILD &&
             BendPoitr[bi+1]->Status == BENDBUF_FILD &&
             BendPoitr[bi  ]->FrstCode == 0xE0 &&
             BendPoitr[bi  ]->BendCents != 0.0 &&
             BendPoitr[bi+1]->FrstCode != 0x90 &&
             BendPoitr[bi  ]->MidiChannel == BendPoitr[bi+1]->MidiChannel)
         {
            BendPoitr[bi]->Status = BENDBUF_FREE;
            removedEn++;
         }
      }
   } while (removedEn);
}

/***********************************************************************/
void AdhocInitBuffer (void)
{
   int i;

   for (i = 0; i < BENDBUFSIZE; i++)
   {
      BendPoitr[i] = &BendEntry[i];
      BendPoitr[i]->Status = BENDBUF_FREE;
   }
   Tonica = 60;
}

/***********************************************************************/
static int DeDouble (int grado, int *p_NumEltsFilled)
{
   int ind;

   for (ind = 0; ind < NotesInChord; ind++)
   {
      if (Chord[ind] == grado)
         break;
      if (Chord[ind] > 24) /* That is, 1000, signalling an unused slot. */
      {
         (*p_NumEltsFilled)++;
         break;
      }
   }
   return ind;
}

/***********************************************************************/
static int EnoughInChordForNow (
   double SigmaWas, double SigmaNow, int NumEltsFilled)
{
   int quarters;

   for (quarters = 1; quarters < 3; quarters++)
      if (SigmaNow - SigmaWas >= quarters - 0.001 &&
          SigmaNow - SigmaWas <= quarters + 0.1   &&
          /* 4, if this element were accepted. But it is rejected
             for now, so three suffices. */
          NumEltsFilled >= 4)
         return 1;
   if (NumEltsFilled >= NotesInChord)
      return 1;
   if (SigmaNow - SigmaWas >= 4 - 0.001)
      return 1;
   return 0;
}

/***********************************************************************/
void AdhocFillChordArray (int verbose)
{
   int BufIndex, BufPrevx, ChordIndex, NumEltsFilled;

   /* High grade (much higher than 14 for example) marks unused chord
      position. */
   for (ChordIndex = 0; ChordIndex < NotesInChord; ChordIndex++)
      Chord[ChordIndex] = 1000;

   for (BufIndex = 0, NumEltsFilled = 0;
        BufIndex < BENDBUFSIZE && BendPoitr[BufIndex]->Status != BENDBUF_EOF;
        BufIndex++)
   {
      BufPrevx = -1;
      if (BufIndex >= 1 &&
          BendPoitr[BufIndex-1]->FrstCode == 0xE0 &&
          BendPoitr[BufIndex-1]->MidiChannel ==
          BendPoitr[BufIndex  ]->MidiChannel)
         BufPrevx = BufIndex-1;

      if (BendPoitr[BufIndex]->FrstCode == 0x90)
      {
         double gradodbl;
         double bendcents = 0;
         /* Add 240, to ensure a positive result, 0..1 */
         int grado = (BendPoitr[BufIndex]->MidiNote - Tonica + 240) % 12;

         BendPoitr[BufIndex]->Grado12 = grado;
         if (BufPrevx >= 0)
         {
            bendcents = BendPoitr[BufPrevx]->BendCents;
            BendPoitr[BufPrevx]->Grado12 = grado;
         }
         /* Support for any quarternotes, indicated by bends 50 cents
            up or down. */
         gradodbl = (double)grado + bendcents / 100.0;
         grado = (int)(gradodbl * 2 + 0.5);

         ChordIndex = DeDouble(grado, &NumEltsFilled);
         if (ChordIndex >= NotesInChord)
            break;
         if (EnoughInChordForNow(
               BendPoitr[0]->SigmaQ, BendPoitr[BufIndex]->SigmaQ, NumEltsFilled))
            break;

         Chord[ChordIndex] = grado;
         BendPoitr[BufIndex]->CorrespChordIndex = ChordIndex;
         BendPoitr[BufIndex]->Status = BENDBUF_2HDL;

         if (BufPrevx >= 0)
         {
            BendPoitr[BufPrevx]->CorrespChordIndex = ChordIndex;
            BendPoitr[BufPrevx]->Status = BENDBUF_2HDL;
         }
      }
   }

   if (verbose > 2)
   {
      printf("Chord array contents:");
      for (int i = 0; i < NotesInChord && Chord[i] < 1000; i++)
         printf(" %2d", Chord[i]);
      printf("\n");
   }
}

/***********************************************************************/
void AdhocRunAdhocAlgorithm (int verbose)
{
   AdhocRunIt(Chord, NotesInChord, wrkC, resC, verbose);
}

/***********************************************************************/
static int TrackCompare (const void *va, const void *vb)
{
   struct rhmidi_eventdata *const *pa = va;
   struct rhmidi_eventdata *const *pb = vb;
   struct rhmidi_eventdata *a = *pa;
   struct rhmidi_eventdata *b = *pb;

   if (b->EndOfTrack && a->EndOfTrack) return  -1;
   if (a->EndOfTrack) return  1;
   if (b->EndOfTrack) return -1;
   if (b->SigmaT == a->SigmaT)
      return a->TrackFilepos > b->TrackFilepos ? 1 : -1;
   return a->SigmaT > b->SigmaT ? 1 : -1;
}

/***********************************************************************/
static int BendBufCompare (const void *va, const void *vb)
{
   struct MidiBendBufData *const *pa = va;
   struct MidiBendBufData *const *pb = vb;
   struct MidiBendBufData *a = *pa;
   struct MidiBendBufData *b = *pb;

   /* Efficiency, avoiding indeterminate values, which quicksort
      isn't good at handling. Also produces identical debug info
      despite different implementations of qsort in the C library.
      Then return the pointer difference is evertything else is
      the same. */
   if ((a->Status == BENDBUF_FREE && b->Status == BENDBUF_FREE) ||
       (a->Status == BENDBUF_EOF  && b->Status == BENDBUF_EOF ) ||
       (a->Status == BENDBUF_2HDL && b->Status == BENDBUF_2HDL))
       return a - b;

   if (a->Status != BENDBUF_FILD || b->Status != BENDBUF_FILD)
      return a->Status - b->Status;

   if      (a->SigmaT      > b->SigmaT)      return +1;
   else if (a->SigmaT      < b->SigmaT)      return -1;
   if      (a->MidiChannel > b->MidiChannel) return +1;
   else if (a->MidiChannel < b->MidiChannel) return -1;
   if      (a->FrstCode    > b->FrstCode)    return -1;
   return +1;
}

/***********************************************************************
   Find a free slot in the Event buffer. Where exactly the new event will
   be placed is immaterial, because the buffer will be sorted before its
   used anyway. See qsort and comparison function BendBufCompare.

   Well, in factor the buffer itself (array of structs) isn't sorted,
   but an array of pointers to its element. More efficient and faster.
 ***********************************************************************/
static int AdhocWhichSlotToUse (void)
{
   static int BufIndex = 0;
   int tried;

   for (tried = 0; tried < BENDBUFSIZE; tried++)
   {
      if  (BendPoitr[BufIndex]->Status != BENDBUF_FILD)
         break;
      else
         BufIndex++;
      if (BufIndex >= BENDBUFSIZE)
         BufIndex = 0;
   }
   return tried >= BENDBUFSIZE ? -1 : BufIndex;
}

/***********************************************************************/
static void AdhocFillBenddata (int BufIndex, struct rhmidi_eventdata *EvtdatTrack)
{
   if (EvtdatTrack->EndOfTrack)
   {
      BendPoitr[BufIndex]->Status   = BENDBUF_EOF;
      return;
   }
   else
      BendPoitr[BufIndex]->Status   = BENDBUF_FILD;

   BendPoitr[BufIndex]->FrstCode    = EvtdatTrack->FrstCode;
   BendPoitr[BufIndex]->MidiChannel = EvtdatTrack->ScndCode;
   BendPoitr[BufIndex]->SigmaT      = EvtdatTrack->SigmaT;
   BendPoitr[BufIndex]->SigmaQ      = EvtdatTrack->SigmaQ;

   if (EvtdatTrack->FrstCode == 0xE0)
   {
      /* 'Bend' event */
      BendPoitr[BufIndex]->BendCents    = CnvFmMidiBend(EvtdatTrack->EventData);
      BendPoitr[BufIndex]->EventFilepos = EvtdatTrack->TrackFilepos - 2;
   }
   else if (EvtdatTrack->FrstCode == 0x90)
   {
      /* 'Note on' event */
      BendPoitr[BufIndex]->MidiNote	= EvtdatTrack->EventData[0];
   }
}

/***********************************************************************/
static int PickupTonality (struct rhmidi_eventdata *EvtdatTrack)
{
   /* Multiply by 7, reason: "kwintencirkel",  circle of fifth.
      Add 240 to always get a positive result in 0..11 */
   int tonica = (((signed int)EvtdatTrack->EventData[0] * 7) + 240) % 12;

   /* Account for minor/major. See  "FF 59 02" in
      http://www.music.mcgill.ca/~ich/classes/mumt306/StandardMIDIfileformat.html */
   tonica = tonica - (EvtdatTrack->EventData[1] ? 3 : 0) + 60;

   return tonica;
}

/***********************************************************************/
static void ShowBuf (int progline)
{
   fprintf(stdout, "\nProgram line %4d in %s\n", progline, __FILE__);
   for (int i = 0; i < BENDBUFSIZE; i++)
   {
      fprintf(stdout,
         "%2d: status %3d sigmaQ %5.2f code %02x channel %2d note %2x %3d, bend %3.0f\n", i,
         BendPoitr[i]->Status,
         BendPoitr[i]->SigmaQ,
         BendPoitr[i]->FrstCode,
         BendPoitr[i]->MidiChannel,
         BendPoitr[i]->MidiNote,
         BendPoitr[i]->MidiNote,
         BendPoitr[i]->BendCents);
   }
}

static long FileposBendTrack;
static long BytesInBendTrack;
static LLI  PrevSigmaT;
/***********************************************************************/
void AdhocInitliseBendTrack (FILE *fpo, long WhereNumTracks, int NumTracks)
{
   char buf[2];

   /* Increment the track counter in the header, because we're going
      to add a track. */
   fseek(fpo, WhereNumTracks, SEEK_SET);
   CnvToHighByteFirst(buf, 2, NumTracks + 1);
   fwrite(buf, 2, 1, fpo);

   /* Announce the track */
   fseek(fpo, 0, SEEK_END);
   fwrite("MTrk", 1, 4, fpo);
   FileposBendTrack = ftell(fpo);
   fwrite("\x00\x00\x00\x00", 1, 4, fpo);
   BytesInBendTrack = 0;
   PrevSigmaT = 0;
}

/***********************************************************************/
void AdhocFinaliseBendTrack (FILE *fpo)
{
   char lenbuf[4];

   CnvToHighByteFirst(lenbuf, 4, BytesInBendTrack + 4);
   fseek(fpo, FileposBendTrack, SEEK_SET);
   fwrite(lenbuf, 1, 4, fpo);
   fseek(fpo, 0, SEEK_END);
   fwrite("\x00\xFF\x2F\x00", 1, 4, fpo);
}

/***********************************************************************/
int AdhocWriteBendEvents (FILE *fpo)
{
   int BufIndex = 0;

   for (BufIndex = 0;
        BufIndex < BENDBUFSIZE;
        BufIndex++)
   {
      if (BendPoitr[BufIndex]->Status == BENDBUF_EOF)
         return EOF;
      if (BendPoitr[BufIndex]->Status != BENDBUF_2HDL)
         break;

      int ind = BendPoitr[BufIndex]->CorrespChordIndex;
      double bendcents = rats2deviafrmequi(
         resC[ind].Considered.numer, resC[ind].Considered.denom,
         BendPoitr[BufIndex]->Grado12);
      char BendBuf[2];

      if (BendPoitr[BufIndex]->FrstCode == 0xE0)
      {
         CnvToMidiBend(BendBuf, bendcents);
         fseek(fpo, BendPoitr[BufIndex]->EventFilepos, SEEK_SET);
         fwrite(BendBuf, 1, 2, fpo);

         /* Mark this event as handled, and the slot as free. */
         BendPoitr[BufIndex]->Status = BENDBUF_FREE;
         /* Ignore next note-on event, because the modified existing
            bend already covers it. */
         BufIndex++;
      }
      else /* Could only be 0x90 then */
      {
         unsigned char DeltaTbuf[4];
         int Deltabytes;
         unsigned char StatusChannel = 0xE0;

         Deltabytes = CnvToVariableLengthValue(DeltaTbuf,
            BendPoitr[BufIndex]->SigmaT - PrevSigmaT);
         fseek(fpo, 0, SEEK_END);
         fwrite(DeltaTbuf, 1, Deltabytes, fpo);

         StatusChannel = 0xE0 + BendPoitr[BufIndex]->MidiChannel;
         fwrite(&StatusChannel, 1, 1, fpo);
         CnvToMidiBend(BendBuf, bendcents);
         fwrite(BendBuf, 1, 2, fpo);

         PrevSigmaT = BendPoitr[BufIndex]->SigmaT;
         BytesInBendTrack += Deltabytes + 1 + 2;
      }
      BendPoitr[BufIndex]->Status = BENDBUF_FREE;
   }
   return 0;
}
